@@ -1,3 +1,4 @@
+import time
 import os
 import json
 import base64
@@ -51,6 +52,9 @@ class RESTInterface:
         self._api_key_secret = config["api_key_secret"]
         self._pretty = config["pretty"]
         self._dry_run = config["dry_run"]
+        self._retry_count = config["retry_count"]
+        self._retry_delay = config["retry_delay"]
+        self._retry_aggressive = config["retry_aggressive"]
 
         auth_header_value = "Bearer " + base64.b64encode(
             bytes(f"{self._api_key_id}:{self._api_key_secret}", "utf-8")
@@ -91,45 +95,83 @@ class RESTInterface:
         else:
             raise Exception(response_data)
 
+    def _handle_retries(self, func, retry_count=None):
+        if retry_count is None:
+            retry_count = self._retry_count
+        for i in range(retry_count + 1):
+            try:
+                return func()
+            except Exception as e:
+                if not self._retry_aggressive:
+                    lqs_expected_error_codes = [
+                        "[BadRequest]",
+                        "[Forbidden]",
+                        "[NotFound]",
+                        "[Conflict]",
+                        "[Locked]",
+                    ]
+                    for code in lqs_expected_error_codes:
+                        if code in str(e):
+                            raise e
+                if retry_count > 0 and i < retry_count:
+                    # exponential backoff
+                    backoff = self._retry_delay * (2**i)
+                    logger.error(f"Error: {e}")
+                    logger.info(f"Retrying in {backoff} seconds")
+                    time.sleep(backoff)
+                else:
+                    raise e
+
     def _get_resource(self, resource_path):
         if self._dry_run:
             logger.info(f"(Dry Run) GET {self._api_url}/{resource_path}")
             return {}
-        r = requests.get(f"{self._api_url}/{resource_path}", headers=self._headers)
-        try:
-            response_data = r.json()
-        except json.decoder.JSONDecodeError:
-            raise Exception(f"Error: {r.text}")
-        if r.ok:
-            return response_data
-        else:
-            raise Exception(response_data)
+
+        def make_request():
+            r = requests.get(f"{self._api_url}/{resource_path}", headers=self._headers)
+            return self._handle_response_data(r)
+
+        return self._handle_retries(make_request)
 
     def _create_resource(self, resource_path, data):
         if self._dry_run:
             logger.info(f"(Dry Run) POST {self._api_url}/{resource_path}, {data}")
             return {}
-        r = requests.post(
-            f"{self._api_url}/{resource_path}",
-            data=json.dumps(data),
-            headers=self._headers,
-        )
-        return self._handle_response_data(r)
+
+        def make_request():
+            r = requests.post(
+                f"{self._api_url}/{resource_path}",
+                data=json.dumps(data),
+                headers=self._headers,
+            )
+            return self._handle_response_data(r)
+
+        return self._handle_retries(make_request)
 
     def _update_resource(self, resource_path, data):
         if self._dry_run:
             logger.info(f"(Dry Run) PATCH {self._api_url}/{resource_path}, {data}")
             return {}
-        r = requests.patch(
-            f"{self._api_url}/{resource_path}",
-            data=json.dumps(data),
-            headers=self._headers,
-        )
-        return self._handle_response_data(r)
+
+        def make_request():
+            r = requests.patch(
+                f"{self._api_url}/{resource_path}",
+                data=json.dumps(data),
+                headers=self._headers,
+            )
+            return self._handle_response_data(r)
+
+        return self._handle_retries(make_request)
 
     def _delete_resource(self, resource_path):
         if self._dry_run:
             logger.info(f"(Dry Run) DELETE {self._api_url}/{resource_path}")
             return
-        r = requests.delete(f"{self._api_url}/{resource_path}", headers=self._headers)
-        return self._handle_response_data(r)
+
+        def make_request():
+            r = requests.delete(
+                f"{self._api_url}/{resource_path}", headers=self._headers
+            )
+            return self._handle_response_data(r)
+
+        return self._handle_retries(make_request)
