@@ -108,7 +108,7 @@ class Node:
             offset += limit
             # fetch message data
             for record in records:
-                # record["bytes"] = requests.get(record["bytes_url"]).content
+                record["bytes"] = None  # init to none to avoid race condition in publish check
                 threading.Thread(target=self.fetch_record_bytes, args=(record,)).start()
             # combine records, sort by timestamp
             self._subscriptions[topic_name]["records"] += records
