@@ -1,7 +1,16 @@
+import os
+import time
+import logging
 from uuid import UUID
 from enum import Enum
 from typing import Optional, Union, List
 
+logging.basicConfig(
+    level=os.getenv("LQS_LOG_LEVEL") or logging.INFO,
+    format="%(asctime)s  (%(levelname)s - %(name)s): %(message)s",
+)
+logger = logging.getLogger(__name__)
+
 import requests
 import xmltodict
 from urllib.parse import unquote
@@ -32,9 +41,41 @@ class S3Resource(str, Enum):
 
 
 class S3:
-    def __init__(self, creator):
+    def __init__(self, config, creator):
+        self._config = config
         self._creator = creator
 
+        self._retry_count = config["retry_count"]
+        self._retry_delay = config["retry_delay"]
+        self._retry_aggressive = config["retry_aggressive"]
+
+    def _handle_retries(self, func, retry_count=None):
+        if retry_count is None:
+            retry_count = self._retry_count
+        for i in range(retry_count + 1):
+            try:
+                return func()
+            except Exception as e:
+                if not self._retry_aggressive:
+                    lqs_expected_error_codes = [
+                        "[BadRequest]",
+                        "[Forbidden]",
+                        "[NotFound]",
+                        "[Conflict]",
+                        "[Locked]",
+                    ]
+                    for code in lqs_expected_error_codes:
+                        if code in str(e):
+                            raise e
+                if retry_count > 0 and i < retry_count:
+                    # exponential backoff
+                    backoff = self._retry_delay * (2**i)
+                    logger.error(f"Error: {e}")
+                    logger.info(f"Retrying in {backoff} seconds")
+                    time.sleep(backoff)
+                else:
+                    raise e
+
     def generate_presigned_url(
         self,
         method: S3Method,
@@ -95,15 +136,20 @@ class S3:
             "Key": key,
             "UploadId": upload_id,
         }
-        url, used_params = self.generate_presigned_url(
-            method="abort_multipart_upload",
-            resource=resource,
-            params=params,
-            resource_id=resource_id,
-            timestamp=timestamp,
-        )
-        r = requests.delete(url)
-        return r.headers, used_params, None
+
+        def make_requests():
+            url, used_params = self.generate_presigned_url(
+                method="abort_multipart_upload",
+                resource=resource,
+                params=params,
+                resource_id=resource_id,
+                timestamp=timestamp,
+            )
+
+            r = requests.delete(url)
+            return r.headers, used_params, None
+
+        return self._handle_retries(make_requests)
 
     def complete_multipart_upload(
         self,
@@ -132,13 +178,7 @@ class S3:
             "Key": key,
             "UploadId": upload_id,
         }
-        url, used_params = self.generate_presigned_url(
-            method="complete_multipart_upload",
-            resource=resource,
-            params=params,
-            resource_id=resource_id,
-            timestamp=timestamp,
-        )
+
         complete_multipart_payload = "<CompleteMultipartUpload>"
         for part in parts:
             if "ETag" not in part or "PartNumber" not in part:
@@ -152,15 +192,25 @@ class S3:
             """
         complete_multipart_payload += "</CompleteMultipartUpload>"
 
-        r = requests.post(url, data=complete_multipart_payload)
-        if decode_response:
-            response_text = unquote(r.text)
-        else:
-            response_text = r.text
-        response_content = (
-            xmltodict.parse(response_text) if convert_to_dict else response_text
-        )
-        return r.headers, used_params, response_content
+        def make_requests():
+            url, used_params = self.generate_presigned_url(
+                method="complete_multipart_upload",
+                resource=resource,
+                params=params,
+                resource_id=resource_id,
+                timestamp=timestamp,
+            )
+            r = requests.post(url, data=complete_multipart_payload)
+            if decode_response:
+                response_text = unquote(r.text)
+            else:
+                response_text = r.text
+            response_content = (
+                xmltodict.parse(response_text) if convert_to_dict else response_text
+            )
+            return r.headers, used_params, response_content
+
+        return self._handle_retries(make_requests)
 
     def create_multipart_upload(
         self,
@@ -184,22 +234,26 @@ class S3:
         else:
             timestamp = None
         params = {"Key": key}
-        url, used_params = self.generate_presigned_url(
-            method="create_multipart_upload",
-            resource=resource,
-            params=params,
-            resource_id=resource_id,
-            timestamp=timestamp,
-        )
-        r = requests.post(url)
-        if decode_response:
-            response_text = unquote(r.text)
-        else:
-            response_text = r.text
-        response_content = (
-            xmltodict.parse(response_text) if convert_to_dict else response_text
-        )
-        return r.headers, used_params, response_content
+
+        def make_requests():
+            url, used_params = self.generate_presigned_url(
+                method="create_multipart_upload",
+                resource=resource,
+                params=params,
+                resource_id=resource_id,
+                timestamp=timestamp,
+            )
+            r = requests.post(url)
+            if decode_response:
+                response_text = unquote(r.text)
+            else:
+                response_text = r.text
+            response_content = (
+                xmltodict.parse(response_text) if convert_to_dict else response_text
+            )
+            return r.headers, used_params, response_content
+
+        return self._handle_retries(make_requests)
 
     def delete_object(
         self,
@@ -224,15 +278,19 @@ class S3:
         params = {"Key": key}
         if version_id is not None:
             params["VersionId"] = version_id
-        url, used_params = self.generate_presigned_url(
-            method="delete_object",
-            resource=resource,
-            params=params,
-            resource_id=resource_id,
-            timestamp=timestamp,
-        )
-        r = requests.delete(url)
-        return r.headers, used_params, None
+
+        def make_requests():
+            url, used_params = self.generate_presigned_url(
+                method="delete_object",
+                resource=resource,
+                params=params,
+                resource_id=resource_id,
+                timestamp=timestamp,
+            )
+            r = requests.delete(url)
+            return r.headers, used_params, None
+
+        return self._handle_retries(make_requests)
 
     def get_object(
         self,
@@ -257,15 +315,19 @@ class S3:
         params = {"Key": key}
         if part_number is not None:
             params["PartNumber"] = part_number
-        url, used_params = self.generate_presigned_url(
-            method="get_object",
-            resource=resource,
-            params=params,
-            resource_id=resource_id,
-            timestamp=timestamp,
-        )
-        r = requests.get(url)
-        return r.headers, used_params, r.content
+
+        def make_requests():
+            url, used_params = self.generate_presigned_url(
+                method="get_object",
+                resource=resource,
+                params=params,
+                resource_id=resource_id,
+                timestamp=timestamp,
+            )
+            r = requests.get(url)
+            return r.headers, used_params, r.content
+
+        return self._handle_retries(make_requests)
 
     def head_object(
         self,
@@ -293,17 +355,21 @@ class S3:
             params["PartNumber"] = part_number
         if version_id is not None:
             params["VersionId"] = version_id
-        url, used_params = self.generate_presigned_url(
-            method="head_object",
-            resource=resource,
-            params=params,
-            resource_id=resource_id,
-            timestamp=timestamp,
-        )
-        r = requests.head(url)
-        # TODO: may need to return a status code as well
-        # for now, we'll just return it as the body
-        return r.headers, used_params, {"exists": r.status_code == 200}
+
+        def make_requests():
+            url, used_params = self.generate_presigned_url(
+                method="head_object",
+                resource=resource,
+                params=params,
+                resource_id=resource_id,
+                timestamp=timestamp,
+            )
+            r = requests.head(url)
+            # TODO: may need to return a status code as well
+            # for now, we'll just return it as the body
+            return r.headers, used_params, {"exists": r.status_code == 200}
+
+        return self._handle_retries(make_requests)
 
     def list_multipart_uploads(
         self,
@@ -326,22 +392,26 @@ class S3:
         else:
             timestamp = None
         params = {}
-        url, used_params = self.generate_presigned_url(
-            method="list_multipart_uploads",
-            resource=resource,
-            params=params,
-            resource_id=resource_id,
-            timestamp=timestamp,
-        )
-        r = requests.get(url)
-        if decode_response:
-            response_text = unquote(r.text)
-        else:
-            response_text = r.text
-        response_content = (
-            xmltodict.parse(response_text) if convert_to_dict else response_text
-        )
-        return r.headers, used_params, response_content
+
+        def make_requests():
+            url, used_params = self.generate_presigned_url(
+                method="list_multipart_uploads",
+                resource=resource,
+                params=params,
+                resource_id=resource_id,
+                timestamp=timestamp,
+            )
+            r = requests.get(url)
+            if decode_response:
+                response_text = unquote(r.text)
+            else:
+                response_text = r.text
+            response_content = (
+                xmltodict.parse(response_text) if convert_to_dict else response_text
+            )
+            return r.headers, used_params, response_content
+
+        return self._handle_retries(make_requests)
 
     def list_object_versions(
         self,
@@ -364,22 +434,26 @@ class S3:
         else:
             timestamp = None
         params = {}
-        url, used_params = self.generate_presigned_url(
-            method="list_object_versions",
-            resource=resource,
-            params=params,
-            resource_id=resource_id,
-            timestamp=timestamp,
-        )
-        r = requests.get(url)
-        if decode_response:
-            response_text = unquote(r.text)
-        else:
-            response_text = r.text
-        response_content = (
-            xmltodict.parse(response_text) if convert_to_dict else response_text
-        )
-        return r.headers, used_params, response_content
+
+        def make_requests():
+            url, used_params = self.generate_presigned_url(
+                method="list_object_versions",
+                resource=resource,
+                params=params,
+                resource_id=resource_id,
+                timestamp=timestamp,
+            )
+            r = requests.get(url)
+            if decode_response:
+                response_text = unquote(r.text)
+            else:
+                response_text = r.text
+            response_content = (
+                xmltodict.parse(response_text) if convert_to_dict else response_text
+            )
+            return r.headers, used_params, response_content
+
+        return self._handle_retries(make_requests)
 
     def list_objects_v2(
         self,
@@ -405,22 +479,26 @@ class S3:
         params = {}
         if continuation_token is not None:
             params["ContinuationToken"] = continuation_token
-        url, used_params = self.generate_presigned_url(
-            method="list_objects_v2",
-            resource=resource,
-            params=params,
-            resource_id=resource_id,
-            timestamp=timestamp,
-        )
-        r = requests.get(url)
-        if decode_response:
-            response_text = unquote(r.text)
-        else:
-            response_text = r.text
-        response_content = (
-            xmltodict.parse(response_text) if convert_to_dict else response_text
-        )
-        return r.headers, used_params, response_content
+
+        def make_requests():
+            url, used_params = self.generate_presigned_url(
+                method="list_objects_v2",
+                resource=resource,
+                params=params,
+                resource_id=resource_id,
+                timestamp=timestamp,
+            )
+            r = requests.get(url)
+            if decode_response:
+                response_text = unquote(r.text)
+            else:
+                response_text = r.text
+            response_content = (
+                xmltodict.parse(response_text) if convert_to_dict else response_text
+            )
+            return r.headers, used_params, response_content
+
+        return self._handle_retries(make_requests)
 
     def list_parts(
         self,
@@ -448,22 +526,26 @@ class S3:
         params = {"Key": key, "UploadId": upload_id}
         if part_number_marker is not None:
             params["PartNumberMarker"] = part_number_marker
-        url, used_params = self.generate_presigned_url(
-            method="list_parts",
-            resource=resource,
-            params=params,
-            resource_id=resource_id,
-            timestamp=timestamp,
-        )
-        r = requests.get(url)
-        if decode_response:
-            response_text = unquote(r.text)
-        else:
-            response_text = r.text
-        response_content = (
-            xmltodict.parse(response_text) if convert_to_dict else response_text
-        )
-        return r.headers, used_params, response_content
+
+        def make_requests():
+            url, used_params = self.generate_presigned_url(
+                method="list_parts",
+                resource=resource,
+                params=params,
+                resource_id=resource_id,
+                timestamp=timestamp,
+            )
+            r = requests.get(url)
+            if decode_response:
+                response_text = unquote(r.text)
+            else:
+                response_text = r.text
+            response_content = (
+                xmltodict.parse(response_text) if convert_to_dict else response_text
+            )
+            return r.headers, used_params, response_content
+
+        return self._handle_retries(make_requests)
 
     def put_object(
         self, resource: S3Resource, resource_id: Union[str, dict], key: str, body: bytes
@@ -482,15 +564,19 @@ class S3:
         else:
             timestamp = None
         params = {"Key": key}
-        url, used_params = self.generate_presigned_url(
-            method="put_object",
-            resource=resource,
-            params=params,
-            resource_id=resource_id,
-            timestamp=timestamp,
-        )
-        r = requests.put(url, data=body)
-        return r.headers, used_params, None
+
+        def make_requests():
+            url, used_params = self.generate_presigned_url(
+                method="put_object",
+                resource=resource,
+                params=params,
+                resource_id=resource_id,
+                timestamp=timestamp,
+            )
+            r = requests.put(url, data=body)
+            return r.headers, used_params, None
+
+        return self._handle_retries(make_requests)
 
     def restore_object(self):
         raise NotImplementedError("Restore object is not implemented yet.")
@@ -518,12 +604,16 @@ class S3:
         else:
             timestamp = None
         params = {"Key": key, "PartNumber": part_number, "UploadId": upload_id}
-        url, used_params = self.generate_presigned_url(
-            method="upload_part",
-            resource=resource,
-            params=params,
-            resource_id=resource_id,
-            timestamp=timestamp,
-        )
-        r = requests.put(url, data=body)
-        return r.headers, used_params, None
+
+        def make_requests():
+            url, used_params = self.generate_presigned_url(
+                method="upload_part",
+                resource=resource,
+                params=params,
+                resource_id=resource_id,
+                timestamp=timestamp,
+            )
+            r = requests.put(url, data=body)
+            return r.headers, used_params, None
+
+        return self._handle_retries(make_requests)
